import os
import subprocess

import numpy as np
from scipy.spatial.distance import squareform
from tqdm import tqdm
import quadpy


scheme = quadpy.sphere.lebedev_131()


num_to_element = [0, 'H', 'He', 'Li', 'Be', 'B', 'C', 'N', 'O', 'F', 'Ne', 'Na', 'Mg', 'Al', 'Si', 'P', 'S',
                  'Cl', 'Ar', 'K', 'Ca', 'Sc', 'Ti', 'V', 'Cr', 'Mn', 'Fe', 'Co',
                  'Ni', 'Cu', 'Zn', 'Ga', 'Ge', 'As', 'Se', 'Br', 'Kr', 'Rb', 'Sr', 'Y', 'Zr', 'Nb', 'Mo', 'Tc',
                  'Ru', 'Rh', 'Pd', 'Ag', 'Cd', 'In', 'Sn', 'Sb', 'Te',
                  'I', 'Xe', 'Cs', 'Ba', 'La', 'Ce', 'Pr', 'Nd', 'Pm', 'Sm', 'Eu', 'Gd', 'Tb', 'Dy', 'Ho', 'Er',
                  'Tm', 'Yb', 'Lu', 'Hf', 'Ta', 'W', 'Re', 'Os', 'Ir',
                  'Pt', 'Au', 'Hg', 'Tl', 'Pb', 'Bi', 'Po', 'At', 'Rn', 'Fr', 'Ra', 'Ac', 'Th', 'Pa', 'U', 'Np',
                  'Pu', 'Am', 'Cm', 'Bk', 'Cf', 'Es', 'Fm', 'Md', 'No',
                  'Lr', 'Rf', 'Db', 'Sg', 'Bh', 'Hs', 'Mt', 'Ds', 'Rg', 'Cn', 'Uut', 'Uuq', 'Uup', 'Uuh', 'Uus',
                  'Uuo']


class DistanceGeometry:
    def __init__(self, dgsol, distances, elements, vectorized_input=True):
        self.dgsol = dgsol
        self.elements = elements
        self.distances = self.vector_to_square(distances) if vectorized_input else distances
        self.coords = []

    def vector_to_square(self, vectorized_distances):
        return [squareform(vec_dist) for vec_dist in vectorized_distances]

    def gen_cerror_overview(self):
        print('Error Type, Min, Mean, Max')
        print(f'minError: {np.min(self.c_errors[:, 1])}, {np.mean(self.c_errors[:, 1])}, {np.max(self.c_errors[:, 1])}')
        print(f'avgError: {np.min(self.c_errors[:, 2])}, {np.mean(self.c_errors[:, 2])}, {np.max(self.c_errors[:, 2])}')
        print(f'maxError: {np.min(self.c_errors[:, 2])}, {np.mean(self.c_errors[:, 3])}, {np.max(self.c_errors[:, 3])}')

    def to_scientific_notation(self, number):
        a, b = '{:.17E}'.format(number).split('E')
        num = '{:.12f}E{:+03d}'.format(float(a) / 10, int(b) + 1)
        return num[1:]

    def solve_distance_geometry(self, outpath, n_solutions=10):
        construction_errors = []

        for i, ids in tqdm(enumerate(np.arange(len(self.distances))), total=len(self.distances)):
            out = f'{outpath}/{ids:04}'
            os.makedirs(out, exist_ok=True)
            self.write_dgsol_input(distances=self.distances[i], outpath=out)
            self.solve_dgp(out, n_solutions=n_solutions)
            errors = self.parse_dgsol_errors(out)
            lowest_errors_idx = np.argsort(errors[:, 2])
            construction_errors.append(errors[lowest_errors_idx[0]])
            self.output_to_xyz(out,
                               solution_id=lowest_errors_idx[0],
                               n_solutions=n_solutions,
                               elements=self.elements[i])
        self.c_errors = np.array(construction_errors)

    def write_dgsol_input(self, distances, outpath):
        n, m = np.triu_indices(distances.shape[1], k=1)
        with open(f'{outpath}/dgsol.input', 'w') as outfile:
            for i, j in zip(n, m):
                outfile.write(
                    f'{i + 1:9.0f}{j + 1:10.0f}   {self.to_scientific_notation(distances[i, j])}   '
                    f'{self.to_scientific_notation(distances[i, j])}\n')

    def solve_dgp(self, outpath, n_solutions=10):
        cmd = f'{self.dgsol} -s{n_solutions} {outpath}/dgsol.input {outpath}/dgsol.output {outpath}/dgsol.summary'
        process = subprocess.Popen(cmd.split(), stdout=subprocess.PIPE)
        output, error = process.communicate()
        if error is not None:
            print(outpath, error)

    def parse_dgsol_errors(self, outpath):
        """
        There are 4 types of errors in the dgsol output:

        f_err      The value of the merit function
        derr_min      The smallest error in the distances
        derr_avg      The average error in the distances
        derr_max      The largest error in the distances
        """
        with open(f'{outpath}/dgsol.summary', 'r') as input:
            lines = input.readlines()

        errors = []
        # skip the header lines
        for line in lines[5:]:
            errors.append(line.split()[2:])   # the first two entries are n_atoms and n_distances
        return np.array(errors).astype('float32')

    def output_to_xyz(self, outpath, solution_id, n_solutions, elements):
        n_atoms = elements.shape[0]
        with open(f'{outpath}/dgsol.output') as outfile:
            lines = outfile.readlines()

        coords = []
        for line in lines:
            if not line.startswith('\n') and len(line) > 30:
                coords.append([float(n) for n in line.split()])
        coords = np.array(coords).reshape(n_solutions, n_atoms, 3)
        self.coords.append(coords[solution_id])

        with open(f'{outpath}/dgsol.xyz', 'w') as outfile:
            outfile.write(f'{n_atoms}\n')
            outfile.write('\n')
            for xyz, element in zip(coords[solution_id], elements):
                outfile.write(f'{num_to_element[int(element)]}\t{xyz[0]}\t{xyz[1]}\t{xyz[2]}\n')


def lebedev_sphere_opt(bond, points_distances, origin):
    """
    Draws a Lebedev sphere around a central atom.
    Assumes that the bonded atom is in the origin

    Parameters
    ----------
    bond: float
        Bond length of heavy atom to hydrogen bond.
    points_distances: tuple
        Contains positions (coordinates) of atoms and distances of atoms to the hydrogen in question.
    origin: np.array
        Coordinates of the central heavy atom.

    Returns
    -------
    candidate_position: np.array
        Coordinates of the hydrogen position with the lowest distance error to given points.

    """
    candidate_positions = scheme.points * bond
    deltas = np.sum(
        [(np.linalg.norm(candidate_positions - (pos - origin), axis=1) - dist) ** 2 for pos, dist in points_distances],
        axis=0)
    return candidate_positions[np.argmin(deltas)] + origin


def get_hydrogen_positions(positions, distances, n_hydrogens):
    """
    Calculates hydrogen positions around a given central atom.
    Can return a maximum number of 3 hydrogen positions.

    Parameters
    ----------
    positions: np.array, shape(4, 3)
        Heavy atom coordinates of the closest 4 heavy atoms in the hydrogen environment.
    distances: np.array, shape(5)
        Distances to the hydrogen in investion. First 4 distances correspond to heavy atom hydrogen distances.
        The last distance is the distance between two hydrogens in case multiple hydrogens are being attached.
    n_hydrogens: int
        Number of hydrogens to attach.

    Returns
    -------
    pos_hydr_1: np.array, shape(3)
        Coordinates of a hydrogen atom.
    pos_hydr_2: np.array, shape(3)
        Coordinates of a hydrogen atom. Is None when non-existent.
    pos_hydr_3: np.array, shape(3)
        Coordinates of a hydrogen atom. Is None when non-existent.
    """
    h_h_distance = distances[-1]
    pos_1 = positions[0]
    dist_1 = distances[0]
    pos_hydr_1 = lebedev_sphere_opt(bond=dist_1, origin=pos_1, points_distances=zip(positions[1:], distances[1:-1]))
    if n_hydrogens > 1:
        p_dist_pairs = list(zip(positions[1:-1], distances[1:-2])) if len(positions) > 2 else list(zip(positions[1], distances[1]))
        p_dist_pairs.append((pos_hydr_1, h_h_distance))
        pos_hydr_2 = lebedev_sphere_opt(bond=dist_1, origin=pos_1, points_distances=p_dist_pairs)
    else:
        pos_hydr_2 = None
    if n_hydrogens == 3:
        p_dist_pairs = [(positions[1], distances[1]), (pos_hydr_1, h_h_distance), (pos_hydr_2, h_h_distance)]
        pos_hydr_3 = lebedev_sphere_opt(bond=dist_1, origin=pos_1, points_distances=p_dist_pairs)
    else:
        pos_hydr_3 = None
    return pos_hydr_1, pos_hydr_2, pos_hydr_3


def hydrogen_lebedev_reconstruction(heavy_atom_coords, predicted_h_distances, heavy_atom_hydrogen_mapping):
    """
    Main function to start hydrogen reconstruction.
    Uses a Lebedev optimization scheme to determine hydrogen coordinates.

    Parameters
    ----------
    heavy_atom_coords: np.array, shape(n_atoms, 3)
        Heavy atom coordinates.
    predicted_h_distances: np.array, shape(n_hyrogens, 5)
        Predicted hydrogen distances of each hydrogen.
    heavy_atom_hydrogen_mapping: tuple
        Contains mapping indices for (n_hydrogens_total, (central_heavy_atom, n_hydrogens_on_atom, neighbor_indices)).

    Returns
    -------
    hydrogen_coords: np.array, shape(n_hydrogens, 3)
        Calculated hydrogen coordinates.

    """
    hydrogen_coords = []
    for i in range(len(heavy_atom_hydrogen_mapping)):
        n_hydrogens = len(heavy_atom_hydrogen_mapping[i][1])
        nbh_idxs = heavy_atom_hydrogen_mapping[i][2]
        center_atom_idx = heavy_atom_hydrogen_mapping[i][0]
        pred_positions = heavy_atom_coords[np.array([center_atom_idx, *nbh_idxs])]
        pred_h_distances = predicted_h_distances[i]

        pos_hydr_1, pos_hydr_2, pos_hydr_3 = get_hydrogen_positions(pred_positions, pred_h_distances,
                                                                    n_hydrogens=n_hydrogens)
        hydrogen_coords.append(pos_hydr_1)
        if pos_hydr_2 is not None:
            hydrogen_coords.append(pos_hydr_2)
        if pos_hydr_3 is not None:
            hydrogen_coords.append(pos_hydr_3)
    return np.array(hydrogen_coords)

